/*
 * This is CacheGo, copyright Evan Daniel 2002.
 *
 */

/*
 * This is the actual cache portion of the program.
 * 
 * It allows the engine to make potentially expensive
 * function calls frequently (ie every move it considers)
 * and get cached results, without knowing or caring
 * whether the result has already been computed.
 *
 * The basic idea is that when a function is called,
 * the Cache begins recording what else gets called
 * before it returns (go recursion!).  It keeps track of
 * these, and marks them as the dependencies of the function.
 * The next time the function is called with the same
 * arguments, the cache checks to see if it has a cached result
 * for the function with the same arguments.  If it does,
 * it checks the dependencies.  If the dependencies return the
 * same values, the cached value of the function is called,
 * without actually calling the function itself.
 *
 * The cache also keeps track of the board, for consistency purposes.
 *
 * All the cleverness lies in the dependency tracking.
 * This may need some work to improve performance.
 * Currently, anything that makes no cached calls is assumed
 * to be uncacheable.
 *
 * Ideally, we would have no uncacheables, just a set of functions
 * that return the state of the board, for which we would invalidate
 * the relevant cache entries when trying or making a move.
 *
 * FIXME:
 * Allow more than one cached result per function call.
 * Better uncacheables handling.
 * Dependency tracking needs an overhaul before multiple
 * results per call can happen.  Basically, it needs to be a
 * tree instead of a list that stores the dependencies of a
 * result.
 */

import java.lang.reflect.*;
import java.lang.*;
import java.util.*;

class Cache
{
	private Engine myEngine;
	private Hashtable cache;
	public FunctionResult depPtr;
	Cache(Engine e)
	{
		System.out.println("Intializing cache.");
		myEngine = e;
		cache = new Hashtable();
		depPtr = null;
	}

	public Object call(String name, Object[] params)
	{
		Class[] paramTypes = new Class[params.length];
		for (int  i = 0; i < params.length; i++)
		{
			paramTypes[i] = params[i].getClass();
		}

		Method m;

		try
		{
			m = myEngine.getClass().getMethod(name, paramTypes);
		} catch (NoSuchMethodException e)
		{
			System.err.println("Couldn't find method " + name);
			System.err.println(e.toString());
			return null;
		}
		FunctionCall f = new FunctionCall(m, params);
		return call(f);
	}

	public Object call(FunctionCall fCall)
	{
		//return the result of FunctionCall fCall,
		//hopefully from the cache.
		//If it's not in cache, add it.
		
		//update dependencies
		if (depPtr != null)
		{
			depPtr.addDep(fCall);
		}

		//next, check if our copy is out of date

		if (!cache.containsKey(fCall))
		{
			//we don't have it.  call it for real.
			System.out.println("No cached result for " + fCall.function.getName());
			FunctionResult r =  new FunctionResult(myEngine, this, fCall);
			cache.put(fCall, r);
			return r.retval;
		}

		//ok, we have one.  validate, return as needed.

		if (validate(fCall))
		{
			System.out.println("Valid cached result for " + fCall.function.getName());
			FunctionResult r = ((FunctionResult)(cache.get(fCall)));
			return r.retval;
		} else
		{
			System.out.println("Invalid cached result for " + fCall.function.getName());
			FunctionResult r =  new FunctionResult(myEngine, this, fCall);
			cache.put(fCall, r);
			return r.retval;
		}
	}

	private boolean validate(FunctionCall fCall)
	{
		//return true if cached value was valid,
		//false if we had to change it.
		FunctionResult oldResult = ((FunctionResult)(cache.get(fCall)));
		if (oldResult == null)
		{
			cache.put(fCall, new FunctionResult(myEngine, this, fCall));
			return false;
		}
		
		if (!cacheable(fCall))
		{	
			FunctionResult newResult = new FunctionResult(myEngine, this, fCall);
			cache.put(fCall, newResult);
			return oldResult.equals(newResult);
		} else
		{
			boolean depsValid = true;
			for (int i = 0; i < oldResult.deps.size(); i++)
			{
				depsValid = depsValid && validate(((FunctionCall)(oldResult.deps.elementAt(i))));
			}

			if (depsValid)
			{
				return true;
			} else
			{
				FunctionResult newResult = new FunctionResult(myEngine, this, fCall);
				cache.put(fCall, newResult);
				return oldResult.equals(newResult);
			}
		}
	}
	
	private boolean cacheable(FunctionCall fCall)
	{
		FunctionResult r = ((FunctionResult)(cache.get(fCall)));
		//this should never happen.
		if (r == null) return false;
		return r.deps.size() != 0;
	}

}
