/*
 * This is CacheGo, copyright Evan Daniel 2002.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation - version 2.     
 *                                                             
 * This program is distributed in the hope that it will be       
 * useful, but WITHOUT ANY WARRANTY; without even the implied    
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       
 * PURPOSE.  See the GNU General Public License in file COPYING  
 * for more details.                                             
                                                              
 * You should have received a copy of the GNU General Public     
 * License along with this program; if not, write to the Free    
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       
 * Boston, MA 02111, USA.                                        
 *
 */

/*
 * This is the GTP interface portion.
 * Basic functionality is as follows:
 * Create a GTP object, with 3 io streams.  It will parse
 * from the input stream until it recieves the quit command.
 * Functions are automatically found by introspection,
 * so to create a new GTP function, just create the java
 * function with the same name preceded by gtp_.
 * These functions can take any number of parameters of any type,
 * provided that type has a constructor that takes a single string.
 * (The autoconstructing code to implement this has not yet been written).
 * Typical types are board coordinates and Integers (not ints).
 *
 * FIXME:
 * Still a couple bugs in regexs.
 * Write autoconstructors for function parameters.
 * Improved error checking (and recovery?).
 * Remove debug code to output valid GTP.
 * Change debug code over to java.util.logging stuff.
 */


import java.lang.*;
import java.lang.reflect.*;
import java.util.regex.*;
import java.io.*;

class GTP
{
	private BufferedReader in;
	private PrintStream out;
	private PrintStream err;
	private String copyright = "This program copyright Evan Daniel, 2002";
	private String id = "";
	private Engine myEngine;
	
	
	public static void main(String args[])
	{
		//create a gtp thing, run as needed
		GTP mGTP = new GTP(System.in, System.out, System.err);
	}

	GTP(InputStream rawIn, PrintStream rawOut, PrintStream rawErr)
	{
		//create a new gtp interface that reads from in,
		//writes to out, and errors to err.
		//main loop is in here.  this should probably be turned
		//into a separate thread.
		out = rawOut;
		err = rawErr;
		
		//so we can read lines at a time
		in = new BufferedReader(new InputStreamReader(rawIn));
		
		err.println(copyright);

		boolean done = false;

		while(!done)
		{
			done = parseGTP();
		}
	}

	protected boolean parseGTP()
	{
		//returns true if GTP command was quit, false
		//otherwise.  Calls the function specified by the GTP
		//command with the specified arguments.
		
		String command = "";
		String function;
		String paramString;
		String[] params;
		id = "";
		//the preparsing pattern -- optional whitespace and
		//trailing comment.
		//reluctantly match everything up to the comment,
		//so that the comment gets matched when possible.
		Pattern preParse = Pattern.compile("\\s*(.*?)(#.*)?");
		Matcher preMatcher;
		while(command == "")
		{
			//read a line, throw away comments and blanks
			try
			{
				command = in.readLine();
			} catch (IOException e)
			{
				System.err.println("Error reading GTP stream:");
				System.err.println(e.toString());
			}

 			preMatcher = preParse.matcher(command);
			if (!preMatcher.matches()) System.err.println("GTP Preparser failed to match.  This should never happen.");
			
			//the command (not whitespace or comment) portion
			command = preMatcher.group(1);
		}
		
		//the gtp command matching pattern:
		//an optional id, a function name, and optionally
		//some parameters.
		//note that doubled backslashes are required to get past the
		//Java string parser -- the string literal has double
		//backslashes, the string passed to the Pattern does not.
		Pattern p = Pattern.compile("\\s*(\\d*)\\s*(\\S+)\\s*(\\s(.*))?");
		Matcher m = p.matcher(command);
		m.matches();
		if (m.groupCount() == 1)
		{
			function = m.group(1);
		}
		id = m.group(1);
		function = m.group(2);
		paramString = m.group(4);//not yet split up.
		
		//FIXME: split up parameter string.
		params = new String[1];
		params[0] = paramString;

		if (function.equals("quit"))
		{
			//true if done (ie command was quit)
			return true;
		}
		
		function = "gtp_" + function;
		//now we have a function name.  Look up the corresponding
		//Method, and figure out how to call it.
		
		Class myClass = this.getClass();
		Method[] GTPMethods = myClass.getDeclaredMethods();
		boolean found = false;
		//System.out.println("Looking for Method " + function);
		for (int i = 0; i < GTPMethods.length; i++)
		{
			//for each Method, check if it's the one we want.
			//If so, try to figure out how to call it.
			Method curMethod = GTPMethods[i];
			//System.out.println("Found Method " + curMethod.getName());
			if (curMethod.getName().equals(function))
			{
				//it's got the right name
				if (curMethod.getReturnType() != Void.TYPE)
				{
					System.err.println("GTP function " + curMethod.getName() + " is not a void function and therefore invalid.");
					continue;
				}
				//it's got the right name and return type

				/*
				if (curMethod.getParameterTypes() != {String.getClass()})
				{
					System.err.println("GTP function " + curMethod.getName() + " does not take a String as its only formal parameter.");
					continue;
				}
				*/

				//ok, we can call it now.
				try
				{
					curMethod.invoke(this, params);
				} catch (Exception e)
				{
					//oops
					System.err.println("Error calling GTP function " + curMethod.getName() + ":");
					System.err.println(e.toString());
					System.err.println(e.getCause().toString());
				}
			}
		}

		//it wasn't a quit command
		return false;
	}

	protected void beginGTPSuccess()
	{
		out.print("=" + id + " ");
	}

	protected void beginGTPFailure()
	{
		out.print("?" + id + " ");
	}

	protected void beginGTP(boolean success)
	{
		if (success)
		{
			beginGTPSuccess();
		} else
		{
			beginGTPFailure();
		}
	}

	protected void finishGTP()
	{
		out.print("\n\n");
	}
	
	protected void returnGTP(boolean success, String result)
	{
		beginGTP(success);
		out.print(result);
		finishGTP();
	}

	private void gtp_echo(String param)
	{
		//System.out.println("Entering echo method.");
		returnGTP(true, param);
	}

	private void gtp_new_engine(String param)
	{
		myEngine = new Engine();
	}
	
	private void gtp_setA(String param)
	{
		System.out.println("'" + param + "'");
		try
		{
			myEngine.setA(Integer.parseInt(param));
		} catch (NumberFormatException e)
		{
			returnGTP(false, e.toString());
			return;
		}
		returnGTP(true, "");
		return;
	}

	private void gtp_getA(String param)
	{
		returnGTP(true, Integer.toString(myEngine.getA()));
		return;
	}
	
	private void gtp_setB(String param)
	{
		System.out.println("'" + param + "'");
		try
		{
			myEngine.setB(Integer.parseInt(param));
		} catch (NumberFormatException e)
		{
			returnGTP(false, e.toString());
			return;
		}
		returnGTP(true, "");
		return;
	}

	private void gtp_getB(String param)
	{
		returnGTP(true, Integer.toString(myEngine.getB()));
		return;
	}

	private void gtp_setC(String param)
	{
		System.out.println("'" + param + "'");
		try
		{
			myEngine.setC(Integer.parseInt(param));
		} catch (NumberFormatException e)
		{
			returnGTP(false, e.toString());
			return;
		}
		returnGTP(true, "");
		return;
	}

	private void gtp_getC(String param)
	{
		returnGTP(true, Integer.toString(myEngine.getC()));
		return;
	}

	private void gtp_sum(String param)
	{
		returnGTP(true, Integer.toString(myEngine.sum()));
		return;
	}

	private void gtp_sum3(String param)
	{
		returnGTP(true, Integer.toString(myEngine.sum3()));
		return;
	}

}
